/*  cond_logdet, a program to estimate conditioned logdet distances from gene content data.
 *  Copyright (C) 2006 by Matthew Spencer
 *
 *  This program is free software; you may redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *  For any other enquiries write to Matthew Spencer, School of Biological Sciences,
 *  University of Liverpool, Crown Street, Liverpool, L69 7ZB, UK.
 *
 *  Or send E-mail to m.spencer@liverpool.ac.uk
 *
 *
--------------------------------------------------------------------------------------------------------------
 *  Parts of this program (functions to read phylip file) are based on fastDNAml by Gary J. Olsen, which is in turn based
 *  partly on dnaml by Joseph Felsenstein
 *
 *  Copyright notice from dnaml:
 *
 *     version 3.3. (c) Copyright 1986, 1990 by the University of Washington
 *     and Joseph Felsenstein.  Written by Joseph Felsenstein.  Permission is
 *     granted to copy and use this program provided no fee is charged for it
 *     and provided that this copyright notice is not removed.
 *
-----------------------------------------------------------------------------------------------------
 *  Uses the Mersenne Twister pseudorandom number generator.  Copyright notice from mt19937ar.c:
 *   A C-program for MT19937, with initialization improved 2002/1/26.
 *  Coded by Takuji Nishimura and Makoto Matsumoto.
 *
 *  Before using, initialize the state by using init_genrand(seed)  
 *  or init_by_array(init_key, key_length).
 *
 *  Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
 *  All rights reserved.                          
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *    3. The names of its contributors may not be used to endorse or promote 
 *       products derived from this software without specific prior written 
 *       permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 *  Any feedback is very welcome.
 *  http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/emt.html
 *  email: m-mat @ math.sci.hiroshima-u.ac.jp (remove space)
 *
-----------------------------------------------------------------------------
 * Uses some code from seq-gen.  Copyright notice for seq-gen:
 *  Sequence Generator - seq-gen, version 1.3.2
 *  Copyright (c)1996-2004, Andrew Rambaut & Nick Grassly
 *  Department of Zoology, University of Oxford			
 *  All rights reserved.                          
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *    3. The names of its contributors may not be used to endorse or promote 
 *       products derived from this software without specific prior written 
 *       permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 *  Any feedback is very welcome.
 *  http://evolve.zoo.ox.ac.uk/software/Seq-Gen/
 *  email: andrew.rambaut@zoo.ox.ac.uk
 *
----------------------------------------------------------------------
 * Incorporates the SGEFA package, for which the READ.ME file is

**********************************************************************
**********************************************************************
**********************************************************************
********************    <<< DISCLAIMER >>>      **********************
**********************************************************************
**********************************************************************
**********************************************************************
This  document was prepared   as an account  of  work sponsored  by an
agency of the  United States  Government.   Neither  the United States
Government  nor the University     of  California nor   any   of their
employees, makes any warranty,  express  or implied,  or assumes   any
legal liability or responsibility for  the accuracy, completeness,  or
usefulness  of   any  information,  apparatus,    product,  or process
disclosed, or represents  that  its use would not   infringe privately
owned  rights.  Reference herein to  any specific commercial products,
process, or    service  by trade   name, trademark,  manufacturer,  or
otherwise, does not  necessarily constitute or  imply its endorsement,
recommendation, or  favoring by  the United States  Government or  the
University of California.  The views and opinions of authors expressed
herein do not necessarily state or reflect those of  the United States
Government  thereof, and shall not be  used for advertising or product
endorsement purposes.
**********************************************************************
**********************************************************************
**********************************************************************

           This is the READ.ME for the SGEFA/SGESL package.

        SGEFA is a   gaussian  elimination (with  partial    pivoting)
package written in C (single precision).  It is a translation from the
FORTRAN LINPACK routines of the same names.  The matrix data structure
was modified  to account for   the conventions of   C.  That  is,  all
elements of the  matrix are referenced  from 0 instead  of  1 and  the
matrix is stored as a series of vectors (which  are the columns).  The
matrix data  structure is defined in ge.h.   The user must supply  the
storage for the columns  of the matrix by  calls to the  standard UNIX
memory allocation routine MALLOC.  See  the matgen routine in driver.c
for  an example.  The necessary BLAS  routines are included in blas.c.
Easy access to the matrix elements can  be had by looking  at the elem
and pelem macros defined in ge.h.

        A sample  output (from a  Sun-3/160   workstation with FPA) is
included in the file sgefat.out

        Send comments, suggestions and bug-reports to:
                Dr. Mark K. Seager
                Lawrence Livermore Nat. Lab.
                PO Box 808, L-316
                Livermore, CA 94550
                (415) 423-3141
                seager@lll-crg.Arpa
----------------------------------------------------------------------------------------
 *
 *
 *LIST OF CHANGES
 * 27 September 2006: incorporated code to read phylip format files from fastDNAml
 * 28 September 2006: incorporated Mersenne Twister code from mt19937ar.c by Takuji Nishimura and Makoto Matsumoto
 * 9 January 2007: added new way of dealing with distances where the determinant is negative: set to Inf.  Changed to non-ansi in Makefile so we can
 * use the math.h definitions of NaN and Inf
 * 18 May 2007: added option to calculate just the determinant of the pattern prob matrix
 * 20 June 2007: added option to estimate distances by an empirical Bayes method when determinant is negative
 * 28 June 2007: added option to use constrained ML + pseudocount estimation method
 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <getopt.h>
#include <ctype.h>
#include <math.h>
#include "mtrand.h"
#include "newton.h"

#define PROGRAM_NAME "cond_logdet"
#define PROGRAM_VERSION "0.3"
#define nmlngth         10  /* number of characters in species name */
#define TRUE             1
#define FALSE            0
#define LEN		1000
typedef int boolean;

/*structure to hold genome data*/
typedef struct
{
	int **gdata;
	int ntaxa;
	int ngenes;
	char **names;
	int *weights;
} genomedata;

/*structure to hold settings*/
typedef struct
{
	int reps;
	boolean quiet;
	int estmethod;
} parameters;

/*structure for linked list of pairs with det F <= 0*/
struct dprob
{
  int i,j; /*pair of genomes*/
  struct dprob* next; /*pointer to next item in list*/
  int *N; /*pattern counts for this pair*/
};

/*estimation options*/
enum{ML,Laplace,CML,DET,EBP,CMP};

void ReadParams(int argc, char **argv, parameters *params, char *infilename, char *outfilename);
void programinfo(parameters *params, genomedata *gd);
double det_twostate(double **A);
double logdet_twostate(double **F, parameters *params, int *n, double *sumd, double *sumd2, struct dprob *current, struct dprob **head,int i,int j, int cgs, int modlist);
double log_statedet_twostate(double **F);
double **make_2darray(int m, int n);
void free_2darray(double **A, int m);
double ***make_3darray(int l, int m, int n);
void free_3darray(double ***A, int l, int m);
genomedata *make_genomedata(int ntaxa, int ngenes);
void free_genomedata(genomedata *gd);
genomedata *read_genomedata(char *filename);
boolean whitechar (int ch);
boolean bindigitchar (int ch);
boolean getdata (genomedata *gd,FILE *infile);
void printgd(genomedata *gd);
void init_weights(genomedata *gd);
void equal_weights(genomedata *gd);
void bootstrap_weights(genomedata *gd);
boolean cond_dists(genomedata *gd, double ***dmcubec, parameters *params);
double get_cond_freq(int s1, int s2, int i, int j, int k, genomedata *gd, parameters *params,int cgs);
boolean get_all_cond_freqs(double **F,int i, int j, int k, genomedata *gd,parameters *params,int cgs);
void print_2darray(double **A, int m, int n);
void print_3darray(double ***A, int l, int m, int n);
void make_cond_phylip(double ***dmcubec, char *fname, genomedata *gd, int rep);
void write_cgsizes(char *fname,genomedata *gd, int rep);
char *makeoutfilename(char *fname, int rep, char *suffix);
int get_cgsize(int k,genomedata *gd);
int main(int argc, char **argv)
{
	genomedata *gd;
	parameters *params;
	int r=1;
	unsigned long randomSeed;
	double ***dmcubec;
	char *infilename, *outfilename;

	/*initialize pseudorandom number generator*/
	randomSeed = CreateSeed();
	init_genrand(randomSeed);

	/*read parameters*/
	params=(parameters *)calloc(1,sizeof(parameters));
	infilename=(char *)calloc(LEN,sizeof(char));
	outfilename=(char *)calloc(LEN,sizeof(char));
	ReadParams(argc,argv,params,infilename,outfilename);

	/*read in the data*/
	gd=read_genomedata(infilename);

	/*show the settings*/
	if(!params->quiet) programinfo(params,gd);

	/*allocate memory for set of all distances for all conditioning genomes*/
	dmcubec=make_3darray(gd->ntaxa,gd->ntaxa,gd->ntaxa);
	if(params->reps==0){/*no bootstrapping*/
		equal_weights(gd);/*all weights set to 1*/
		cond_dists(gd,dmcubec,params);/*get conditioned logdet dists for data, all conditioning genomes*/
		make_cond_phylip(dmcubec,outfilename,gd,0);/*write distances to output file*/
		write_cgsizes(outfilename,gd,0);/*write conditioning genome sizes to second output file*/
	}
	while(r<=params->reps){/*do bootstrap reps*/
		bootstrap_weights(gd);/*generate a set of weights for gene families*/
		if(cond_dists(gd,dmcubec,params)){/*get conditioned logdet dists for bootstrap data, all conditioning genomes*/
			make_cond_phylip(dmcubec,outfilename,gd,r);/*write distances to output file if rep worked*/
			write_cgsizes(outfilename,gd,r);/*write conditioning genome sizes to second output file*/
			if(!params->quiet) printf("rep %d\n",r);
			r++;
		}
		else if(!params->quiet) printf("NaN distances...throwing away rep\n");
	}

	/*clean up*/
	free_3darray(dmcubec,gd->ntaxa,gd->ntaxa);
	free_genomedata(gd);
	free(infilename);
	free(outfilename);
	free(params);

	return EXIT_SUCCESS;
}

/*read parameters*/
void ReadParams(int argc, char **argv, parameters *params, char *infilename, char *outfilename)
{
	int c;
	extern char *optarg;
	params->reps=0;
	params->quiet=TRUE;
	params->estmethod=ML;
	opterr = 0;
	while ((c = getopt (argc, argv, "qne:r:h")) != -1) switch (c) {
		case 'r':/*number of bootstrap replicates*/
			params->reps=atoi(optarg);
			if(params->reps<0){
				fprintf (stderr,"Invalid number of bootstrap reps: %d\n",params->reps);
				exit(EXIT_FAILURE);
			}
		break;
		case 'q':/*quiet operation*/
			params->quiet=TRUE;
		break;
		case 'n':/*noisy operation*/
			params->quiet=FALSE;
		break;
		case 'e':/*pattern probability estimation*/
			if(!strcmp(optarg,"M")) params->estmethod=ML;
			else if(!strcmp(optarg,"P")) params->estmethod=Laplace;
			else if(!strcmp(optarg,"C")) params->estmethod=CML;
			else if(!strcmp(optarg,"D")) params->estmethod=DET;
			/*else if(!strcmp(optarg,"B")) params->estmethod=EBP;*/ /*disable this: we decided it wasn't really the right thing: Ed, 12 September 08. MS, 2/10/08*/
			else if(!strcmp(optarg,"I")) params->estmethod=CMP;
			else{
				fprintf(stderr,"Invalid estimation method: %s\n",optarg);
				exit(EXIT_FAILURE);
			}
		break;
		case 'h':/*help message*/
			printf("\n"); 
			printf(PROGRAM_NAME);
			printf(" ");
			printf(PROGRAM_VERSION);
			printf(": estimation of conditioned logdet distances\n"); 
			printf("Usage: ");
			printf(PROGRAM_NAME);
			printf(" [-h] [-r#] [-q] [-n] [-e ESTIMATION_METHOD] [infile] [outfile] \n");
			printf("Options: \n");
			printf("-h: this help message\n");
			printf("-r#: number of bootstrap replicates.  If 0 or not specified, analyzes the observed data only.\n");
			printf("If any pairwise distances (excluding those to the conditioning genome) are NaN we throw away the bootstrap rep and draw again.\n");
			printf("-q: quiet operation (default)\n");
			printf("-n: noisy operation: copyright notice, settings, progress, etc\n");
			printf("-e ESTIMATION METHOD: estimation method for pairwise pattern probabilities\n");
			printf("options are M=maximum likelihood (default), P=add-one pseudocounts, C=constrained maximum likelihood, D=return determinant of pattern prob matrix instead of logdet distance, I=ML if determinant positive, otherwise constrained ML + pseudocount\n");
			printf("Input file format: a phylip format data matrix.  It uses old code to read phylip format so taxon names must be exactly 10 characters,\n");
			printf("padded with spaces if necessary.  Number of taxa and gene families on the first row.  0 (absence) and 1 (presence) are the only symbols allowed.\n");
			printf("Output file format: [outfile]_cgsizes is a space-separated ASCII file of conditioning genome sizes \n");
			printf("[outfile]_data is an ASCII file containing conditioned logdet distance matrices for each choice of conditioning genome.\n");
			printf("Each distance matrix is in phylip format: number of taxa in first row, then alternating rows of taxon name and distances to each other taxon.\n");
			printf("The row and column corresponding to the conditioning genome are NaN.\n");
			printf("Conditioning genomes are ordered in the same way as rows of the distance matrices.\n");
			printf("If bootstrapping, each bootstrap replicate generates a separate pair of files: [outfile]_rep#_cgsizes, [outfile]_rep#_data\n");
			printf("Output files can be used as input to BIONJ_cond to get a supertree that combines information from all choices of conditioning genome.\n");
			printf("Copyright (C) 2006 Matthew Spencer, School of Biological Sciences, University of Liverpool (m.spencer@liverpool.ac.uk).\n");
			exit(EXIT_SUCCESS);
		break;
		case '?':
		if (isprint (optopt))
			fprintf (stderr, "Unknown option `-%c'.\n", optopt);
		else
			fprintf (stderr,"Unknown option character `\\x%x'.\n",optopt);
			exit(EXIT_FAILURE);
		default:
			abort ();
	}


	if(argc > optind){
		sprintf(infilename,"%s",argv[optind]);
    	}
	else{/* ask for the input-file */
		printf("\nEnter input file : ");
		scanf("%s",infilename);
	}
	if(argc>(optind+1)){
 		sprintf(outfilename,"%s",argv[optind+1]);
 	}
 	else{/* ask for ouput-file.        */
		printf("\nEnter output file : ");
		scanf("%s",outfilename);
	}

}

/*print some information about the program*/
void programinfo(parameters *params, genomedata *gd)
{
	printf("\n");
	printf(PROGRAM_NAME);
	printf(" ");
	printf(PROGRAM_VERSION);
	printf(", a program to estimate conditioned logdet distances from gene content data.\n");
	printf("Copyright (C) 2006 by Matthew Spencer(m.spencer@liverpool.ac.uk)\n");
   	printf("This program is free software; you may redistribute it and/or modify it\n");
	printf("under the terms of the GNU General Public License as published by the Free\n");
   	printf("Software Foundation; either version 2 of the License, or (at your option)\n");
   	printf("any later version.\n");
	printf("This program is distributed in the hope that it will be useful, but\n");
	printf("WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY\n");
   	printf("or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.\n\n");
	if(params->reps>0) printf("%d bootstrap reps\n",params->reps);
	else printf("No bootstrap\n");
	switch(params->estmethod){
	case ML:
	  printf("Maximum  likelihood estimation of pattern probabilities\n");
	  break;
	case Laplace:
	  printf("Laplace (add-one) pseudocount estimation of pattern probabilities\n");
	  break;
	case CML:
	  printf("Constrained maximum likelihood estimation of pattern probabilities\n");
	  break;
	case DET:
	  printf("Calculate determinant of pattern probability matrix instead of logdet distance\n");
	  break;
	case EBP:
	  printf("Use ML estimate of pattern probabilities if determinant positive, otherwise use posterior mode of empirical Bayes estimate\n");
	  break;
	case CMP:
	  printf("Use ML estimate of pattern probabilities if determinant positive, otherwise use constrained ML + pseudocount estimate\n");
	  break;
	default:
	  printf("Unknown estimation method\n");
	  break;
	}
	printf("%d taxa, %d gene families\n",gd->ntaxa,gd->ngenes);
}

/*two-state logdet distance*/
/*n, sumd and sumd2 point to variables that will hold sample size, sum and sum of squares for determinants in
  cases where determinant>0. The two dprob structures are for a linked list of pairs where there was a problem to fix later
  i and j are the current pair, and cgs is current conditioning genome size. If modlist is zero, don't update n, sumd, sumd2 or the linked list: just get distance*/
double logdet_twostate(double **F, parameters *params, int *n, double *sumd, double *sumd2,struct dprob *current,struct dprob **head,int i,int j,int cgs,int modlist)
{
	double d,fi0,fi1,fj0,fj1,ld;
	
	d=det_twostate(F);
	if(params->estmethod==DET) return d;/*just the determinant of the pattern prob matrix*/
	if(params->estmethod==EBP && modlist!=0){
	  if(d>0){/*use in empirical prior for determinant*/
	    *sumd+=d*4;/*multiply by 4: rescale so range [0,1], then we will get shape parameters for a beta fit to the rescaled data*/
	    *sumd2+=d*d*16;
	    *n+=1;
	  }
	  else{/*add this one to the list of problems to fix later*/
	    current=(struct dprob*)calloc(1,sizeof(struct dprob));
	    if(current==NULL){
	      fprintf(stderr,"Couldn't get space for current pointer to dprob item\n");
	      exit(EXIT_FAILURE);
	    }
	    current->i=i;
	    current->j=j;
	    current->N=(int*)calloc(4,sizeof(int));
	    if(current->N==NULL){
	      fprintf(stderr,"Couldn't get space for pointer to pattern counts in current dprob\n");
	      exit(EXIT_FAILURE);
	    }
	    /*set current pattern counts*/
	    /*we're converting from double to int so need to round not truncate*/
	    current->N[0]=floor(F[0][0] * cgs + 0.5);
	    current->N[1]=floor(F[0][1] * cgs + 0.5);
	    current->N[2]=floor(F[1][0] * cgs + 0.5);
	    current->N[3]=floor(F[1][1] * cgs + 0.5);
	    
	    current->next=*head;
	    *head=current;
	  }
	}
	if(d<0 && params->estmethod==CML)/*conditioned ML is same as ML unless determinant < 0 , in which case it gives Inf distance*/
		return atof("Inf");
	else if(d<=0 && params->estmethod==CMP){/*conditional ML + pseudocount if det <0*/
	  fi0=F[0][0]+F[0][1];
	  fi1=F[1][0]+F[1][1];
	  fj0=F[0][0]+F[1][0];
	  fj1=F[0][1]+F[1][1];

	  F[0][0]=(fi0*fj0*cgs+0.5)/(cgs+1);/*add 1/2 observation pseudocount to independence estimate*/
	  F[0][1]=(fi0*fj1*cgs)/(cgs+1);/*independence estimate*/
	  F[1][0]=(fi1*fj0*cgs)/(cgs+1);/*independence estimate*/
	  F[1][1]=(fi1*fj1*cgs+0.5)/(cgs+1);/*add 1/2 observation pseudocount to independence estimate*/
	  ld= -0.5*(log(det_twostate(F))-0.5*log_statedet_twostate(F));/*logdet for modified F*/
	  if(!params->quiet) printf("Distance i=%d,j=%d estimated as %f by constrained ML + pseudocount\n",i+1,j+1,ld);
	  return ld;

	}
	else
		return -0.5*(log(d)-0.5*log_statedet_twostate(F));
}

/*determinant of a 2x2 matrix*/
double det_twostate(double **A)
{
	return A[0][0] * A[1][1] - A[1][0] * A[0][1];
}

/*log of determinant of product of marginal probabilities from a 2x2 F matrix*/
double log_statedet_twostate(double **F)
{
	double lp0i,lp1i,lp0j,lp1j;

	/*get log marginal probs*/
	lp0i=log(F[0][0]+F[0][1]);
	lp1i=log(F[1][0]+F[1][1]);
	lp0j=log(F[0][0]+F[1][0]);
	lp1j=log(F[0][1]+F[1][1]);
	
	/*get determinant of product of diagonal matrices of marginal probs*/
	return lp0i+lp1i+lp0j+lp1j;

}

/*allocate memory for an mxn array of pointers to double*/
double **make_2darray(int m,int n)
{
	double **A;
	int i;

	A=(double **)calloc(m,sizeof(double*));
	if(A==NULL){
		(void)fprintf(stderr,"ERROR: calloc failed");
		(void)exit(EXIT_FAILURE);
	}

	for(i=0;i<m;i++){
		A[i]=(double*)calloc(n,sizeof(double));
		if(A[i]==NULL){
			(void)fprintf(stderr,"ERROR: calloc failed");
			(void)exit(EXIT_FAILURE);
		}
	}
	return A;
}

/*free memory for mxn array of pointers to double*/
void free_2darray(double **A,int m)
{
	int i;
	for(i=0;i<m;i++){
		free(A[i]);
	}
	free(A);
}

/*allocate memory for an l x m x n array of pointers to double*/
double ***make_3darray(int l,int m,int n)
{
	double ***A;
	int i,j;

	A=(double ***)calloc(l,sizeof(double**));
	if(A==NULL){
		(void)fprintf(stderr,"ERROR: calloc failed");
		(void)exit(EXIT_FAILURE);
	}

	for(i=0;i<l;i++){
		A[i]=(double**)calloc(m,sizeof(double*));
		if(A[i]==NULL){
			(void)fprintf(stderr,"ERROR: calloc failed");
			(void)exit(EXIT_FAILURE);
		}
		for(j=0;j<m;j++){
			A[i][j]=(double*)calloc(n,sizeof(double));
			if(A[i][j]==NULL){
				(void)fprintf(stderr,"ERROR: calloc failed");
				(void)exit(EXIT_FAILURE);
			}
		}
	}
	return A;
}

/*free memory for l x m x n array of pointers to double*/
void free_3darray(double ***A,int l, int m)
{
	int i,j;
	for(i=0;i<l;i++){
		for(j=0;j<m;j++){
			free(A[i][j]);
		}
		free(A[i]);
	}
	free(A);
}


/*read a phylip-format file of gene presence-absence*/
/*return a pointer to a structure containing the data*/
genomedata *read_genomedata(char *filename)
{
	FILE *infile;
	int ntaxa,ngenes;
	genomedata *gd;
	
	/*open the file*/
	if( (infile=fopen(filename,"r")) == NULL)
	{
    		fprintf(stderr,"File not found:%s\n",filename);
    		exit(EXIT_FAILURE);
  	}
	fscanf(infile,"%d%d",&ntaxa,&ngenes);
	gd=make_genomedata(ntaxa,ngenes);

	/*read in the presence/absence data*/
	if (!getdata (gd,infile)){
		fclose(infile);
		exit(EXIT_FAILURE);
	}

	fclose(infile);
	return gd;
}

/*allocate memory for a structure containing genome data*/
genomedata *make_genomedata(int ntaxa, int ngenes)
{
	genomedata *gd;
	int i;

	gd=(genomedata *)calloc(1,sizeof(genomedata));

	gd->gdata=(int **)calloc(ntaxa,sizeof(int*));
	if(gd->gdata==NULL){
		(void)fprintf(stderr,"ERROR: calloc failed");
		(void)exit(EXIT_FAILURE);
	}
	gd->names=(char**)calloc(ntaxa,sizeof(char*));
	if(gd->names==NULL){
		(void)fprintf(stderr,"ERROR: calloc failed");
		(void)exit(EXIT_FAILURE);
	}

	for(i=0;i<ntaxa;i++){
		gd->gdata[i]=(int*)calloc(ngenes,sizeof(int));
		if(gd->gdata[i]==NULL){
			(void)fprintf(stderr,"ERROR: calloc failed");
			(void)exit(EXIT_FAILURE);
		}
		gd->names[i]=(char*)calloc(nmlngth+1,sizeof(char));
		if (gd->names[i]==NULL){
			(void)fprintf(stderr,"ERROR: calloc failed");
			(void)exit(EXIT_FAILURE);
		}
	}
	gd->weights=(int*)calloc(ngenes,sizeof(int));
	gd->ntaxa=ntaxa;
	gd->ngenes=ngenes;
	return gd;
}

/*free memory for structure containing genome data*/
void free_genomedata(genomedata *gd)
{
	int i;
	for(i=0;i<gd->ntaxa;i++){
		free(gd->gdata[i]);
		free(gd->names[i]);
	}
	free(gd->gdata);
	free(gd->names);
	free(gd->weights);
	free(gd);
}

/*read a 0/1 data matrix in phylip format: modified from fastDNAml*/
boolean getdata (genomedata *gd,FILE *infile)
  {
    int   i, j, k, basesread, basesnew, ch;
    char *nameptr;
    boolean  allread, firstpass;


    basesread = basesnew = 0;

    allread = FALSE;
    firstpass = TRUE;
    ch = ' ';

    while (! allread) {
      for (i = 0; i < gd->ntaxa; i++) {     /*  Read data line */

        if (firstpass) {                      /*  Read species names */
          j = 0;
          while (whitechar(ch = getc(infile))) {  /*  Skip blank lines */
            if (ch == '\n')  j = 0;  else  j++;
            }

          if (j > nmlngth) {
            printf("ERROR: Blank name for species %d; ", i+1);
            printf("check number of species,\n");
            printf("       number of gene families, and interleave option.\n");
            return  FALSE;
            }

          nameptr = gd->names[i];
          for (k = 0; k < j; k++)  *nameptr++ = ' ';

          while (ch != '\n' && ch != EOF) {
            if (whitechar(ch))  ch = ' ';
            *nameptr++ = ch;
            if (++j > nmlngth) break;
            ch = getc(infile);
            }

          while (*(--nameptr) == ' ') ;          /*  remove trailing blanks */
          *(++nameptr) = '\0';                   /*  add null termination */

          if (ch == EOF) {
            printf("ERROR: End-of-file in name of species %d\n", i+1);
            return  FALSE;
            }
          }    /* if (firstpass) */

        j = basesread;
        while ((j < gd->ngenes)
               && ((ch = getc(infile)) != EOF)
               && (ch != '\n')) {
          if (bindigitchar(ch)) {
            if(ch=='0')
		gd->gdata[i][j]=0;
            else
		gd->gdata[i][j]=1;
            j++;
            }
          else if (whitechar(ch)) ;
          else {
            printf("ERROR: Bad character (%c) at gene family %d of genome %d\n",
                    ch, j+1, i+1);
            return  FALSE;
            }
          }

        if (ch == EOF) {
          printf("ERROR: End-of-file at gene family %d of genome %d\n", j+1, i+1);
          return  FALSE;
          }

        if (! firstpass && (j == basesread)) i--;        /* no data on line */
        else if (i == 0) basesnew = j;
        else if (j != basesnew) {
          printf("ERROR: genomes out of alignment\n");
          printf("%d (instead of %d) gene families read in genome %d\n",
                  j - basesread, basesnew - basesread, i+1);
          return  FALSE;
          }

        while (ch != '\n' && ch != EOF) ch = getc(infile);  /* flush line */
        }                                                  /* next sequence */
      firstpass = FALSE;
      basesread = basesnew;
      allread = (basesread >= gd->ngenes);
      }
  return TRUE;
  } /* getdata */


boolean whitechar (int ch)
{ 
	return (ch == ' ' || ch == '\n' || ch == '\t');
}

boolean bindigitchar (int ch)
{
	return (ch >= '0' && ch <= '1');
}

/*print out the data*/
void printgd(genomedata *gd)
{
	int i,j;
	printf("%d %d\n",gd->ntaxa,gd->ngenes);
	for(i=0;i<gd->ntaxa;i++){
		printf(gd->names[i]);
		printf("\t");
		for(j=0;j<gd->ngenes;j++){
			printf("%d",gd->gdata[i][j]);
		}
		printf("\n");
	}
	printf("weights\t");
		for(j=0;j<gd->ngenes;j++){
			printf("%d",gd->weights[j]);
		}
	printf("\n\n");
}

/*print contents of 2d array*/
void print_2darray(double **A, int m, int n)
{
	int i,j;
	for(i=0;i<m;i++){
		for(j=0;j<n;j++){
			printf("%f ",A[i][j]);
		}
		printf("\n");
	}
}

/*print contents of 3d array*/
void print_3darray(double ***A, int l, int m, int n)
{
	int i,j,k;
	for(i=0;i<l;i++){
		for(j=0;j<m;j++){
			for(k=0;k<n;k++){
				printf("%f ",A[i][j][k]);
			}
			printf("\n");
		}
		printf("\n");
	}
}


/*set gene family weights to zero*/
void init_weights(genomedata *gd)
{
	int i;
	for(i=0;i<gd->ngenes;i++){
		gd->weights[i]=0;
	}
}

/*set gene family weights to one*/
void equal_weights(genomedata *gd)
{
	int i;
	for(i=0;i<gd->ngenes;i++){
		gd->weights[i]=1;
	}
}


/*set bootstrap weights*/
void bootstrap_weights(genomedata *gd)
{
	int i,j;
	init_weights(gd);/*reset all weights to zero*/
	for (i=0;i<gd->ngenes;i++){
		j=mtrandint(gd->ngenes);/*generate ngenes random integers between 0 and ngenes-1*/
		gd->weights[j]++;/*increment the weight of the corresponding gene*/
	}
}

/*get conditioned logdet dists using bootstrap weights for all conditioning genomes*/
/*returns FALSE if any distance is NaN*/
boolean cond_dists(genomedata *gd, double ***dmcubec,parameters *params)
{
	int i,j,k;
	double **F;
	int n;/*number of distances for empirical Bayes*/
	int cgs;/*size of conditioning genome*/
	double sumd,sumd2;/*these will hold sum of determinants and determinants^2 for empirical Bayes (note they are rescaled to range [0,1])*/
	double mean,var,v,w,t;/*parameters for beta distribution of determinants*/
	struct dprob *current, *head;/*use in linked list of pairs with det F <=0, if using empirical Bayes*/
	const float tolG=1e-6;
	const float tolF=1e-6;
	const int niter=1000;
	struct conv_info *cstatus;
	float *Fvec;

	F=make_2darray(2,2);
	cstatus =(struct conv_info*)calloc(1,sizeof(struct conv_info));
	Fvec    = (float *)calloc(5,sizeof( float ) );/*fifth element of Fvec is Lagrange multiplier lambda*/
	if(cstatus==NULL || Fvec==NULL){
	  fprintf(stderr,"Couldn't get space for cstatus/Fvec\n");
	  exit(EXIT_FAILURE);
	}

	for(k=0;k<gd->ntaxa;k++){/*for each conditioning genome k*/
	  if(!params->quiet) printf("Conditioning genome %d:\n",k+1);
	  head=NULL;
	  current=NULL;
	  sumd=0.;
	  sumd2=0.;
	  n=0;
	  cgs=get_cgsize(k,gd);/*size of this conditioning genome*/
	  for(i=0;i<gd->ntaxa;i++){
	    for(j=i+1;j<gd->ntaxa;j++){/*for each pair of genomes i,j*/
	      if(i==k||j==k)/*distance not defined*/
		dmcubec[i][j][k]=atof("NaN");
	      else{
		if(!get_all_cond_freqs(F,i,j,k,gd,params,cgs)){
		  fprintf(stderr,"ERROR: estimation method not implemented yet\n");
		  exit(EXIT_FAILURE);
		}
		dmcubec[i][j][k]=logdet_twostate(F,params,&n,&sumd,&sumd2,current,&head,i,j,cgs,1);/*get conditioned logdet distance for each pair of genomes*/
		if(!params->quiet && isnan(dmcubec[i][j][k])) printf("Distance i=%d, j=%d not yet estimated\n",i+1,j+1);
		if(params->reps>0 && isnan(dmcubec[i][j][k]) && params->estmethod!=EBP){
		  free_2darray(F,2);
		  return FALSE;/*throw away this rep if bootstrapping and distance is NaN and not using empirical Bayes*/
		}
	      }
	      dmcubec[j][i][k]=dmcubec[i][j][k];/*symmetrical distance*/
	    }
	   }
	   if(params->estmethod==EBP){/*go back and fix the distances with non-positive determinant using empirical Bayes*/
	     mean=sumd/n;/*mean*/
	     var=(sumd2-sumd*mean)/(n-1);/*variance*/

	     /*method of moments estimates of shape parameters for beta distribution*/
	     t=mean*(1-mean)/var-1;
	     v=mean*t;
	     w=(1-mean)*t;
	     current=head;
	     while(current){
	       if(n>1){

		 if(!params->quiet) printf("Distance i=%d,j=%d with conditioning genome %d. N=[%d %d; %d %d], v=%f, w= %f. ",current->i+1,current->j+1,k+1,current->N[0],current->N[1],current->N[2],current->N[3],v,w);

		 FNewton(Fvec,current->N,v,w,tolG,tolF,niter,cstatus);/*get posterior mode estimate of F*/
		 if(cstatus->xstatus==ctolg || cstatus->xstatus==ctolf || cstatus->xstatus==ctolboth){/*converged*/
		   /*get new distance from this F*/
		   F[0][0]=Fvec[0];
		   F[0][1]=Fvec[1];
		   F[1][0]=Fvec[2];
		   F[1][1]=Fvec[3];
		   
		   dmcubec[current->i][current->j][k]=logdet_twostate(F,params,&n,&sumd,&sumd2,current,&head,current->i,current->j,cgs,0);/*get conditioned logdet distance, without modifying n, sumd, sumd2 or list of problem cases*/
		   dmcubec[current->j][current->i][k]=dmcubec[current->i][current->j][k];/*symmetrical distance*/		 
		   if(!params->quiet) printf("Estimated as %f using empirical determinant prior\n",dmcubec[current->i][current->j][k]);
		 }
		 else{/*did not converge: leave distance as nan/inf and show a warning*/
		   if(!params->quiet) printf("Did not converge\n");
		 }
	       }
	       else{/*too few observations to get prior: leave distance as nan/inf and show a warning*/
		 if(!params->quiet) printf("Too few observations to estimate empirical determinant prior\n");
	       }

	       free(current->N);
	       head=current->next;
	       free(current);
	       current=head;
	     }
	   }
	}

	free_2darray(F,2);
	free(cstatus);
	free(Fvec);
	return TRUE;
}

/*count the number of occurrences of states s1 and s2 in genomes i and j conditional on presence in conditioning genome k*/
/*cgs is size of conditioning genome*/
double get_cond_freq(int s1, int s2, int i, int j, int k, genomedata *gd,parameters *params,int cgs)
{
	int g,c;
	c=0;
	for(g=0;g<gd->ngenes;g++){
	  if(gd->gdata[k][g]==1 && gd->gdata[i][g]==s1 && gd->gdata[j][g]==s2) c+=gd->weights[g];
	}

	switch(params->estmethod){
		case ML: case CML: case DET: case EBP: case CMP:/*constrained ML handled in the same way as ML because it always gives Inf distance if on boundary; constrained ML + pseudocount is fixed later if det F <0*/
			return(double)c/cgs;
		break;
		case Laplace:
			return(double)(c+1)/(cgs+4);
		break;
		default:
			return atof("NaN");/*didn't define behaviour for other methods yet*/
		break;
	}
}

/*get size of a conditioning genome*/
int get_cgsize(int k,genomedata *gd)
{
  int g,n;
  n=0;
  for(g=0;g<gd->ngenes;g++){
    if(gd->gdata[k][g]==1) n+=gd->weights[g];
  }
  return n;
}

/*estimate pattern probabilities*/
/*pattern probs F, genomes i and j, conditioning genome k, data in gd, parameters in params, cgs is conditioning genome size*/
boolean get_all_cond_freqs(double **F,int i, int j, int k, genomedata *gd,parameters *params,int cgs)
{
	int s1,s2;
	switch(params->estmethod){
		case ML: case Laplace: case CML: case DET: case EBP: case CMP:/*we can do these based on single-pattern counts*/
			for(s1=0;s1<2;s1++){
				for(s2=0;s2<2;s2++)
					F[s1][s2]=get_cond_freq(s1,s2,i,j,k,gd,params,cgs);/*build the conditioned F matrix using weights*/
			}
		break;
		default:
			return FALSE;/*other methods may later require something that depends on all four pattern counts*/
		break;
	}
	return TRUE;
}


/*make a phylip file containing all the conditioned distance matrices from one rep*/
void make_cond_phylip(double ***dmcubec, char *fname, genomedata *gd, int rep)
{
	FILE *outfile;
	char *outfilename;
	char *suffix;
	int i,j,k;

	/*construct the output file name for this replicate*/
	suffix="_data";
	outfilename=makeoutfilename(fname,rep,suffix);
	if( (outfile=fopen(outfilename,"w")) == NULL)
	{
    		fprintf(stderr,"Couldn't write to %s\n",outfilename);
    		exit(EXIT_FAILURE);
  	}

	for(k=0;k<gd->ntaxa;k++){
		fprintf(outfile,"%d\n",gd->ntaxa);
		for(i=0;i<gd->ntaxa;i++){
			fprintf(outfile,gd->names[i]);
			fprintf(outfile,"\n");
			for(j=0;j<gd->ntaxa;j++){
				fprintf(outfile,"%1.10f ",dmcubec[i][j][k]);
			}
			fprintf(outfile,"\n");
		}
		fprintf(outfile,"\n");
	}
	fclose(outfile);
	free(outfilename);
}

/*write conditioning genome sizes*/
void write_cgsizes(char *fname,genomedata *gd, int rep)
{
	FILE *outfile;
	char *outfilename;
	char *suffix;
	int n,k;

	/*construct the output file name for this replicate*/
	suffix="_cgsizes";
	outfilename=makeoutfilename(fname,rep,suffix);
	if( (outfile=fopen(outfilename,"w")) == NULL)
	{
    		fprintf(stderr,"Couldn't write to %s\n",outfilename);
    		exit(EXIT_FAILURE);
  	}
	for(k=0;k<gd->ntaxa;k++){
	  n=get_cgsize(k,gd);
	  fprintf(outfile,"%d ",n);
	}
	fclose(outfile);
	free(outfilename);
}

/*build output file name*/
/*If rep is 0, we don't append _rep# to the filename*/
char *makeoutfilename(char *fname, int rep, char *suffix)
{
	char *tempstr;
	char numstr[10];
	if(rep>0){/*separate name for each bootstrap rep*/
		sprintf(numstr,"%d",rep);
		tempstr = (char *)malloc((strlen(fname) + strlen("_rep")+strlen(numstr)+strlen(suffix) + 1)*sizeof(char));
		strcpy(tempstr, fname);
		strcat(tempstr,"_rep");
		strcat(tempstr,numstr);
		strcat(tempstr,suffix);
	}
	else{/*we aren't bootstrapping*/
		tempstr = (char *)malloc((strlen(fname)+strlen(suffix) + 1)*sizeof(char));
		strcpy(tempstr, fname);
		strcat(tempstr,suffix);
	}
	return(tempstr);
}


