/*Newton-Raphson code for cond_logdet.c
*  This program is free software; you may redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *  For any other enquiries write to Matthew Spencer, School of Biological Sciences,
 *  University of Liverpool, Crown Street, Liverpool, L69 7ZB, UK.
 *
 *  Or send E-mail to m.spencer@liverpool.ac.uk
 *
----------------------------------------------------------------------
 * Incorporates the SGEFA package, for which the READ.ME is

**********************************************************************
**********************************************************************
********************    <<< DISCLAIMER >>>      **********************
**********************************************************************
**********************************************************************
**********************************************************************
This  document was prepared   as an account  of  work sponsored  by an
agency of the  United States  Government.   Neither  the United States
Government  nor the University     of  California nor   any   of their
employees, makes any warranty,  express  or implied,  or assumes   any
legal liability or responsibility for  the accuracy, completeness,  or
usefulness  of   any  information,  apparatus,    product,  or process
disclosed, or represents  that  its use would not   infringe privately
owned  rights.  Reference herein to  any specific commercial products,
process, or    service  by trade   name, trademark,  manufacturer,  or
otherwise, does not  necessarily constitute or  imply its endorsement,
recommendation, or  favoring by  the United States  Government or  the
University of California.  The views and opinions of authors expressed
herein do not necessarily state or reflect those of  the United States
Government  thereof, and shall not be  used for advertising or product
endorsement purposes.
**********************************************************************
**********************************************************************
**********************************************************************

           This is the READ.ME for the SGEFA/SGESL package.

        SGEFA is a   gaussian  elimination (with  partial    pivoting)
package written in C (single precision).  It is a translation from the
FORTRAN LINPACK routines of the same names.  The matrix data structure
was modified  to account for   the conventions of   C.  That  is,  all
elements of the  matrix are referenced  from 0 instead  of  1 and  the
matrix is stored as a series of vectors (which  are the columns).  The
matrix data  structure is defined in ge.h.   The user must supply  the
storage for the columns  of the matrix by  calls to the  standard UNIX
memory allocation routine MALLOC.  See  the matgen routine in driver.c
for  an example.  The necessary BLAS  routines are included in blas.c.
Easy access to the matrix elements can  be had by looking  at the elem
and pelem macros defined in ge.h.

        A sample  output (from a  Sun-3/160   workstation with FPA) is
included in the file sgefat.out

        Send comments, suggestions and bug-reports to:
                Dr. Mark K. Seager
                Lawrence Livermore Nat. Lab.
                PO Box 808, L-316
                Livermore, CA 94550
                (415) 423-3141
                seager@lll-crg.Arpa
*/


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "sgefac.h"
#include "newton.h"

int get_matrixspace(struct FULL *J,float **G,int **ipvt,float **deltaF);
void free_matrixspace(struct FULL *A);
void matdump(struct FULL *a, char *head );
void set_Jacobian(struct FULL *J,float *F,float v,float w,int *N);
void fvecdump( float *vec, int len, char *head );
void evalminusG(float *G, float *F,int *N, float v, float w, float *deltaF);
void F_update(float *F,float *deltaF);
struct conv_info check_convergence(float *F,float *deltaF,float *G, float tolG, float tolF);
struct conv_info Newton_case(float *F, int *N, float v, float w, float tolG, float tolF, int niter,int showF);
void tryNewton(float *F,float *testF,int *N,float v, float w,float tolG,float tolF, int niter, struct conv_info *cstatus,int showF);



#define TRUE 1
#define FALSE 0



/*get space for the matrix elements*/
int get_matrixspace(struct FULL *J,float **G,int **ipvt,float **deltaF)
{
  int j;
  J->cd=5;
  J->rd=5;/*5 x 5 Jacobian matrix*/
  for(j=0;j<J->rd;j++){
    J->pd[j]=(float*)malloc(J->cd*sizeof(float));
    if(J->pd[j]==NULL){
      fprintf(stderr,"get_matrixspace failed to get memory\n");
      exit(EXIT_FAILURE);
    }
  }
  *deltaF    = (float *)malloc( J->cd*sizeof( float ) );/*fifth element of F is Lagrange multiplier lambda*/
  *G    = (float *)malloc( J->cd*sizeof( float ) );/*current values of Lagrange equation system*/
  *ipvt = (int *)malloc( J->cd*sizeof( int ) );
    if(*deltaF ==NULL || *G == NULL || *ipvt == NULL) {
      fprintf(stderr,"get_matrixspace Can't get enouph space for vectors...\n");
      exit(EXIT_FAILURE);
    }
 
    return(0);
}

/*free space for the matrix elements*/
void free_matrixspace(struct FULL *A)
{
  int j;
  for(j=0;j<A->rd;j++) free(A->pd[j]);
}

/*set the elements of the Jacobian*/
 void set_Jacobian(struct FULL *J,float *F, float v, float w,int *N)
{
  int i;
  float f00,f01,f10,f11,a,b,d,e,f002,f012,f102,f112,vv,ww;
  int n00,n01,n10,n11;
  /*defining these for convenience when converting the Maple results*/
  f00=F[0];
  f01=F[1];
  f10=F[2];
  f11=F[3];
  n00=N[0];
  n01=N[1];
  n10=N[2];
  n11=N[3];
  d=f00*f11-f01*f10;
  e=1-d;
  a=pow(d,2);
  b=pow(e,2);
  f002=pow(f00,2);
  f012=pow(f01,2);
  f102=pow(f10,2);
  f112=pow(f11,2);
  vv=v-1;
  ww=w-1;

  pelem(J,0,0)=-n00/f002-vv*f112/a-ww*f112/b;
  pelem(J,0,1)=vv*f11/a*f10+ww*f11/b*f10;
  pelem(J,0,2)=vv*f11/a*f01+ww*f11/b*f01;
  pelem(J,0,3)=vv/d-vv*f11/a*f00-ww/e-ww*f11/b*f00;

  pelem(J,1,0)=vv*f11/a*f10+ww*f11/b*f10;
  pelem(J,1,1)=-n01/f012-vv*f102/a-ww*f102/b;
  pelem(J,1,2)=-vv/d-vv*f10/a*f01+ww/e-ww*f10/b*f01;
  pelem(J,1,3)=vv*f10/a*f00+ww*f10/b*f00;
  
  pelem(J,2,0)=vv*f11/a*f01+ww*f11/b*f01;
  pelem(J,2,1)=-vv/d-vv*f10/a*f01+ww/e-ww*f10/b*f01;
  pelem(J,2,2)=-n10/f102-vv*f012/a-ww*f012/b;
  pelem(J,2,3)=vv*f01/a*f00+ww*f01/b*f00;

  pelem(J,3,0)=vv/d-vv*f11/a*f00-ww/e-ww*f11/b*f00;
  pelem(J,3,1)=vv*f10/a*f00+ww*f10/b*f00;
  pelem(J,3,2)=vv*f01/a*f00+ww*f01/b*f00;
  pelem(J,3,3)=-n11/f112-vv*f002/a-ww*f002/b;


  for(i=0;i<4;i++){
    pelem(J,4,i)=1.;
    pelem(J,i,4)=1.;
  }
  pelem(J,4,4)=0.;
}

/*evaluate current Lagrange vector and set deltaF to this*/
void evalminusG(float *G, float *F, int *N, float v, float w, float *deltaF)
{
  float lambda,f00,f01,f10,f11,d,x,n00,n01,n10,n11;
  int i;

  f00=F[0];
  f01=F[1];
  f10=F[2];
  f11=F[3];
  lambda=F[4];
  n00=N[0];
  n01=N[1];
  n10=N[2];
  n11=N[3];

  d=f00*f11-f01*f10;
  x=(v-1)/d-(w-1)/(1-d);
  
  G[0]=-n00/f00-f11*x-lambda;
  G[1]=-n01/f01+f10*x-lambda;
  G[2]=-n10/f10+f01*x-lambda;
  G[3]=-n11/f11-f00*x-lambda;
  G[4]=-f00-f01-f10-f11+1.;

  for(i=0;i<5;i++) deltaF[i]=G[i];
}

/*update F = F + delta F*/
void F_update(float *F,float *deltaF)
{
  int i;
  for(i=0;i<5;i++) F[i]=F[i]+deltaF[i];
}

/*check whether we converged*/
/*returns a conv_info structure with xstatus= xiter if we didn't, ctolg if we converged by G criterion, ctolf if we converged by F criterion,
  ctolboth if we converged by both criteria*/
/*sets Gsum and dFsum fields of conv_info to sums of abs RHS and abs change in parameters*/
struct conv_info check_convergence(float *F,float *deltaF,float *G, float tolG, float tolF)
{
  int i;
  struct conv_info ccheck;
  float d;

  ccheck.Gsum=0.;
  ccheck.dFsum=0.;
  ccheck.xstatus=xiter;
  ccheck.stepsrun=0;
  for(i=0;i<5;i++){
    ccheck.Gsum+=fabs(G[i]);/*sum of absolute values of G: should be zero at solution*/
    ccheck.dFsum+=fabs(deltaF[i]);/*sum of absolute values of changes in F: should be zero at solution*/
  }
  /*did we converge?*/
  if(ccheck.Gsum<=tolG && ccheck.dFsum<=tolF) ccheck.xstatus=ctolboth;
  else if(ccheck.Gsum<=tolG) ccheck.xstatus=ctolg;
  else if(ccheck.dFsum<=tolF) ccheck.xstatus=ctolf;
  else ccheck.xstatus=xiter;

  /*is the solution feasible?*/
  d=F[0]*F[3]-F[1]*F[2];/*determinant of F matrix*/
  if(d>=0.) ccheck.feasible=TRUE;
  else ccheck.feasible=FALSE;
  
  return(ccheck);
}

/*run Newton-Raphson for specified initial conditions*/
/*if showF=1, we print out steps in the calculation for debugging*/
struct conv_info Newton_case(float *F, int *N, float v, float w, float tolG, float tolF, int niter,int showF)
{
  int n,retval;
  float *deltaF, *G;
  int *ipvt;
  struct conv_info cstatus;
  struct FULL J;

  /*allocate J,deltaF,G,ipvt*/
  get_matrixspace(&J,&G,&ipvt,&deltaF);/*allocate space for matrix and vectors*/

  if(showF){
    printf("Initial F estimate\n");
    printFmatrix(F);
  }
  /*iterate until convergence:*/
  cstatus.xstatus=xiter;
  cstatus.Gsum=0.;
  cstatus.dFsum=0.;
  cstatus.feasible=FALSE;
  cstatus.stepsrun=0;
      for(n=0;n<niter;n++){
       set_Jacobian(&J,F,v,w,N);/*set Jacobian*/
       if(showF) matdump(&J,"J is...\n");/*print out the matrix J*/
  
      evalminusG(G,F,N,v,w,deltaF);/*evaluate -G and set deltaF to equal it*/
      if(showF) fvecdump(G,5,"-G is...\n");
      retval=sgefa(&J,ipvt);/*LU factorization of J*/
      if(retval){
	if(showF)printf("Zero Column %d found \n", retval);
	cstatus.xstatus=LUfail;
      break;
      }
    else {/*solve J * deltaF = -G*/
      (void)sgesl(&J,ipvt,deltaF,0);/*deltaF is passed equal to G and written to new value*/
      F_update(F,deltaF);/*update F*/
       if(showF) (void)fvecdump(deltaF,J.rd,"delta F\n");
       if(showF)(void)fvecdump(F,J.rd,"New F\n");
      cstatus=check_convergence(F,deltaF,G,tolG,tolF);/*did we converge?*/
      if(cstatus.xstatus!=xiter) break;
    }
  }
  cstatus.stepsrun=n+1;
  if(showF){
    switch (cstatus.xstatus){
    case LUfail: printf("LU factorization failed\n");break;
    case xiter: printf("%d iterations without convergence\n",niter);break;
    case ctolf: printf("converged in %d iterations, sum of abs change in F<=%f\n",n+1,tolF);break;
    case ctolg: printf("converged in %d iterations, sum of abs(RHS)<= %f\n",n+1,tolG);break;
    case ctolboth: printf("converged in %d iterations, sum of abs change in F<=%f, sum of abs(RHS)<=%f\n",n+1,tolF,tolG);break;
    default: printf("not sure of convergence status\n");
    }
    printf("F estimate\n");
    printFmatrix(F);
  }
  free_matrixspace(&J);/*free allocated space*/
  free(deltaF);
  free(G);
  free(ipvt);
  return(cstatus);/*return convergence info*/
}

/*given N, run Newton-Raphson for several sets of initial conditions and pick the one that worked best*/
void FNewton(float *F,int *N,float v,float w,float tolG,float tolF,int niter,struct conv_info *cstatus)
{
  float *testF;
  
  /*float sumN;*/
  
  /*allocate testF*/

  testF    = (float *)malloc(5*sizeof( float ) );/*fifth element of F is Lagrange multiplier lambda*/
  
  if(testF==NULL){
    printf("Couldn't get space for testF\n");
    exit(EXIT_FAILURE);
  }
  cstatus->Gsum=atof("Inf");
  cstatus->dFsum=atof("Inf");
  cstatus->xstatus=xiter;
  cstatus->feasible=FALSE;
  cstatus->stepsrun=0;

  /*try several different initial conditions*/
  
  /*ML estimates of F: seem to give unfeasible solution in all the test cases I tried*/
  /*sumN=N[0]+N[1]+N[2]+N[3];
  testF[0]=(float)N[0]/sumN;
  testF[1]=(float)N[1]/sumN;
  testF[2]=(float)N[2]/sumN;
  testF[3]=(float)N[3]/sumN;
  testF[4]=-1.;*//*initial choice for Lagrange multiplier*/
  /*printf("Trying ML estimate as initial condition\n");
  tryNewton(F,testF,N,v,w,tolG,tolF,niter,cstatus,0);
  printFmatrix(testF);
  print_convinfo(*cstatus);*/

  /*try arbitrary F with positive determinant*/
  testF[0]=0.3;
  testF[1]=0.2;
  testF[2]=0.2;
  testF[3]=0.3;
  testF[4]=-1.;

  /*printf("Trying F with positive determinant as initial condition\n");*/
  tryNewton(F,testF,N,v,w,tolG,tolF,niter,cstatus,0); 
  /*printFmatrix(testF);
    print_convinfo(*cstatus);*/

  /*another arbitrary F with positive determinant*/
  testF[0]=0.45;
  testF[1]=0.05;
  testF[2]=0.05;
  testF[3]=0.45;
  testF[4]=-1.;
  /*printf("Trying F with positive determinant as initial condition\n");*/
  tryNewton(F,testF,N,v,w,tolG,tolF,niter,cstatus,0); 
  /*printFmatrix(testF);
    print_convinfo(*cstatus);*/
 
  free(testF);
}

/*try Newton-Raphson with given initial conditions
  retain if better than existing estimate on Gsum criterion, and feasible
  sets F and cstatus if new estimate better than existing one
  showF=1 means the steps in the Newton-Raphson are printed out (for debugging)
 */
void tryNewton(float *F,float *testF,int *N,float v, float w,float tolG,float tolF, int niter, struct conv_info *cstatus,int showF)
{
  struct conv_info teststatus;
  int i;
  teststatus=Newton_case(testF,N,v,w,tolG,tolF,niter,showF);  /*call the Newton-Raphson*/
  if(showF) print_convinfo(teststatus);
  if(teststatus.Gsum<cstatus->Gsum && teststatus.feasible==TRUE){
    cstatus->Gsum=teststatus.Gsum;
    cstatus->dFsum=teststatus.dFsum;
    cstatus->feasible=teststatus.feasible;
    cstatus->xstatus=teststatus.xstatus;
    cstatus->stepsrun=teststatus.stepsrun;
    for(i=0;i<5;i++) F[i]=testF[i];
  }
}

void printFmatrix(float *F)
{
  printf("%f %f\n",F[0],F[1]);
  printf("%f %f\n",F[2],F[3]);
  printf("Determinant %f\n",F[0]*F[3]-F[1]*F[2]);
  printf("Sum %f\n",F[0]+F[1]+F[2]+F[3]);
  printf("lambda=%f\n",F[4]);
}

void print_convinfo(struct conv_info cstatus)
{
  printf("Convergence status: %d\n",cstatus.xstatus);
  printf("Gsum: %f\n",cstatus.Gsum);
  printf("dFsum: %f\n",cstatus.dFsum);
  printf("Steps run: %d\n",cstatus.stepsrun);
  if(cstatus.feasible==TRUE) printf("Feasible solution\n");
  else printf("Infeasible solution\n");
}

void matdump(struct FULL *a, char *head )

/*
  This routine prints out a FULL matrix with heading head.
*/
{
  register int i, j;
  int k, jj, ncolmod, ncolrem;

  ncolmod = (a->cd)/6;
  ncolrem = a->cd - ncolmod*6;

  printf("%s\n", head );
  for( i=0;  i<a->rd;  i++) {
    printf("%3d|", i );
    j = 0;
    for( k=0; k<ncolmod; k++ ) {
      if( k ) printf("    ");
      for( jj=0;  jj<6;  jj++, j++) 
	printf("%12.4e", pelem(a,i,j));
      printf("\n");
    }
    
    /* Clean up remainder of this row. */
    for( jj=0;  jj<ncolrem;  jj++, j++) 
      printf("%12.4e", pelem(a,i,j));
    printf("\n");
    }
  printf("\n");
}				/* End of MATDUMP */

void fvecdump( float *vec, int len, char *head )
/*
  This routine prints out a float vector `vec' of length `len' and
  a headding of `head'.
*/
{
  register int i, j, count;
  int lenmod, lenrem;

  lenmod = len/6;
  lenrem = len - lenmod*6;

  printf("%s\n", head );
  count = 0;
  for( j=0; j<lenmod; j++ ) {
    printf("%3d|", count );
    for( i=0; i<6; i++, vec++, count++ )
      printf("%12.4e", *vec );
    printf("\n");
  }
  /* Clean up loop. */
  if( lenrem ) {
    printf("%3d|", count );
    for( i=0; i<lenrem; i++, vec++ )
      printf("%12.4e", *vec );
    printf("\n");
  }
  printf("\n");
}				/* End of FVECDUMP */
